#include "BeQuakeView_gl.h"
#include <stdio.h>
#include <stdlib.h>

//*****************
//BeQuakeView - BGLView subclass
//*****************
BeQuakeView::BeQuakeView(int argc, char **argv, BRect R)
	: BGLView(R, "quakeglview", B_FOLLOW_NONE, 0,	//voodoo change from B_FOLLOW_ALL 
				BGL_RGB | BGL_DEPTH | BGL_DOUBLE)  // to B_FOLLOW_NONE
{
	vargc = argc;
	vargv = argv;
	glx = gly = 0;
	glheight = (GLint)R.Height()+1;
	glwidth = (GLint)R.Width()+1;	
	printf("glheight = %d, glwidth = %d\n", glheight, glwidth);
	iy = ix = 0;
	mouseRect = new BRect(0,0,glwidth,glheight);
	mouseRect->InsetBy(50,50);
	set_mouse_position(glwidth/2, glheight/2);
	be_app->HideCursor();
}

BeQuakeView::~BeQuakeView()
{
	delete_sem(quittingSem);
}

void BeQuakeView::MessageReceived(BMessage *msg)
{	
}


void 
BeQuakeView::BGetMouse(void)
{
	release_sem(syncSem); //just for now, otherwise we can get into a blocking situation
	if(Window()->Lock())
	{
		uint32 buttons;
		BPoint point;
		GetMouse(&point, &buttons, false);
		// maybe add a debug build option?
		//if(buttons != 0)
		//	printf("Button %d\n", buttons);
		
		if(!ix)
		{
			ix = point.x;
			iy = point.y;
		}
		BMouse_pos.dx = (point.x - ix)*3;
		BMouse_pos.dy = (point.y - iy)*3;
		if(!mouseRect->Contains(BPoint(point.x, point.y)))
		{
			ix = mouseRect->Width()/2;
			iy = mouseRect->Height()/2;
			set_mouse_position(ix, iy);
		}
		else
		{
			ix = point.x;
			iy = point.y;
		}
		if(buttons & B_PRIMARY_MOUSE_BUTTON) 
			BMouse_pos.Primary = true;
		else 
			BMouse_pos.Primary = false;
		if(buttons & B_SECONDARY_MOUSE_BUTTON) 
			BMouse_pos.Secondary = true;
		else 
			BMouse_pos.Secondary = false;
		if(buttons & B_TERTIARY_MOUSE_BUTTON) 
			BMouse_pos.Tertiary = true;
		else 
			BMouse_pos.Tertiary = false;
		Window()->Unlock();
	}

	acquire_sem(syncSem); //lock again*/
}

void
BeQuakeView::AttachedToWindow()
{
	BGLView::AttachedToWindow();
	quake_thread = spawn_thread(Qthread_func, "Qthread", B_DISPLAY_PRIORITY, this);
	if ((quittingSem = create_sem(1,"quitting_sem")) < B_OK)
		printf("Error: could not start quittingSem\n");
	if ((syncSem = create_sem(1,"sync_sem")) < B_OK)
		printf("Error: could not start syncSem\n");
	resume_thread(quake_thread);
}


void
BeQuakeView::DetachedFromWindow()
{
	BGLView::DetachedFromWindow();
	long dummy;
	long locks=0;
	
	while (Window()->IsLocked())
	{
		locks++;
		Window()->Unlock();
	};
	acquire_sem(quittingSem);
	wait_for_thread(quake_thread ,&dummy);
	release_sem(quittingSem);
	while (locks--)
		Window()->Lock();
}

int32 BeQuakeView::Qthread_func(void *arg)
{
	BeQuakeView *vw = (BeQuakeView *)arg;
	vw->Go();
	return 0;
}

void BeQuakeView::Go(void)
{
	double newtime, oldtime;
	float time;
	app_info info;  
	
	static quakeparms_t parms;
	//allocate global heap - take half available memory or 16MB whichever greater
	system_info sys_info;
	get_system_info(&sys_info);
	int32 available_mem = (sys_info.max_pages - sys_info.used_pages)*4096;

	if (available_mem < (32*1024*1024)) available_mem = 32*1024*1024;
	parms.memsize = available_mem/2; 
	parms.membase = malloc (parms.memsize);
	if (!parms.membase)
		{
		Sys_Error ("Sorry, not enough memory free... Exiting\n");
		return;
		}
	
	be_app->GetAppInfo(&info);  
	BEntry entry(&info.ref);
	BPath path;
	entry.GetPath(&path);
	path.GetParent(&path);
	parms.basedir = path.Path();

	parms.argc = vargc;
	parms.argv = vargv;

	acquire_sem(syncSem); //make sure keyboard code can't run here
	COM_InitArgv(vargc, vargv);
	printf ("Host_Init\n");
	LockGL();
	Host_Init (&parms);
	UnlockGL();
	release_sem(syncSem);
	
	oldtime = Sys_FloatTime();

    while (acquire_sem_etc(quittingSem,1,B_TIMEOUT,0) == B_NO_ERROR)
    {
	    newtime = Sys_FloatTime();
    	time = newtime - oldtime;
		oldtime = newtime;
	 	LockGL();
		//rudolf: keep syncSem close to Host_Frame so enough CPU time is available for the KB
		// scan routine to ever run... (update for version GL1.02)
 		acquire_sem(syncSem);
	    Host_Frame (time);
	    release_sem(syncSem);
	    UnlockGL();
	    release_sem(quittingSem);
    }
}

void BeQuakeView::KeyDown(const char *bytes, int32 numBytes)
{
}

void BeQuakeView::KeyUp(const char *bytes, int32 numBytes)
{
}

void BeQuakeView::MouseDown(BPoint p)
{
/*	release_sem(syncSem); //just for now, otherwise we can get into a blocking situation
	if(Window()->Lock())
	{
	
		if(B_PRIMARY_MOUSE_BUTTON) 
			BMouse_pos.Primary = true;
		if(B_SECONDARY_MOUSE_BUTTON) 
			BMouse_pos.Secondary = true;
		if(B_TERTIARY_MOUSE_BUTTON) 
			BMouse_pos.Tertiary = true;	
		Window()->Unlock();
	}
	acquire_sem(syncSem); //lock again*/
}

void BeQuakeView::MouseUp(BPoint p)
{
/*	release_sem(syncSem); //just for now, otherwise we can get into a blocking situation
	if(Window()->Lock())
	{
	
		if(B_PRIMARY_MOUSE_BUTTON) 
			BMouse_pos.Primary = false;
		if(B_SECONDARY_MOUSE_BUTTON) 
			BMouse_pos.Secondary = false;
		if(B_TERTIARY_MOUSE_BUTTON) 
			BMouse_pos.Tertiary = false;
	
		Window()->Unlock();
	}
	acquire_sem(syncSem); //lock again*/
}

void BeQuakeView::MouseMoved(BPoint p, uint32 t, const BMessage *msg)
{
}

void BeQuakeView::DispatchKeys(const char *bytes, bool down)
{
}
