/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

//CDEngine class mercilessly stolen from CDButtons

#include "cd_beos_gl.h"

void CDAudio_Play(byte track, qboolean looping)
{
engine->Play((int32)track);
}


void CDAudio_Stop(void)
{
engine->Stop();
}


void CDAudio_Pause(void)
{
engine->Pause();
}


void CDAudio_Resume(void)
{
engine->PlayContinue();
}


void CDAudio_Update(void)
{
}


int CDAudio_Init(void)
{
engine = new CDEngine(CDEngine::FindCDPlayerDevice());
if(engine->IsCDPresent())
		{
		Con_Printf("CD Audio initialized\n");
		initialized = true;
		}
	else
		{
		Con_Printf("No Audio CD in drive\n");
		initialized = false;
		}
}


void CDAudio_Shutdown(void)
{
CDAudio_Stop();
delete engine;
initialized = false;
}

CDEngine::CDEngine(int devicefd)
	:	BHandler("CDEngine"),
		devicefd(devicefd)
{
}

CDEngine::CDEngine(BMessage *message)
	:	BHandler(message),
		devicefd(-1)
{
}

CDEngine::~CDEngine()
{
	if(devicefd >= 0)
		close(devicefd);
}

bool
CDEngine::IsCDPresent()
{
	status_t result = ioctl(devicefd, B_SCSI_SCAN);
	if (result != B_NO_ERROR)
		return false;
	return true;
}

void 
CDEngine::Pause()
{
	// pause the CD
	status_t result = ioctl(devicefd, B_SCSI_PAUSE_AUDIO);
	if (result != B_NO_ERROR) {
		PRINT(("error %s pausing\n", strerror(errno)));
		return;
	}
	playing = false;
}

void 
CDEngine::Play(int32 trackNumber)
{
	scsi_play_track track;

	track.start_track = trackNumber;
	track.start_index = 1;
	track.end_track = trackNumber;
	track.end_index = 1;

	status_t result = ioctl(devicefd, B_SCSI_PLAY_TRACK, &track);
	if (result != B_NO_ERROR) {
		PRINT(("error %s playing track\n", strerror(errno)));
		return;
	}
	playing = true;
}

void
CDEngine::PlayContinue()
{
	// continue after a pause
	status_t result = ioctl(devicefd, B_SCSI_RESUME_AUDIO);
	if (result != B_NO_ERROR) {
		PRINT(("error %s resuming\n", strerror(errno)));
		return;
	}
}

void 
CDEngine::Stop()
{
	// stop a playing CD
	status_t result = ioctl(devicefd, B_SCSI_STOP_AUDIO);
	if (result != B_NO_ERROR) {
		PRINT(("error %s stoping\n", strerror(errno)));
		return;
	}
}

static int
try_dir(const char *directory)
{ 
	BDirectory dir; 
	dir.SetTo(directory); 
	if(dir.InitCheck() != B_NO_ERROR) { 
		return false; 
	} 
	dir.Rewind(); 
	BEntry entry; 
	while(dir.GetNextEntry(&entry) >= 0) { 
		BPath path; 
		const char *name; 
		entry_ref e; 
		
		if(entry.GetPath(&path) != B_NO_ERROR) 
			continue; 
		name = path.Path(); 
		
		
		if(entry.GetRef(&e) != B_NO_ERROR) 
			continue; 

		if(entry.IsDirectory()) { 
			if(strcmp(e.name, "floppy") == 0) 
				continue; // ignore floppy (it is not silent) 
			int devfd = try_dir(name);
			if(devfd >= 0)
				return devfd;
		} 
		else { 
			int devfd; 
			device_geometry g; 

			if(strcmp(e.name, "raw") != 0) 
				continue; // ignore partitions 

			devfd = open(name, O_RDONLY); 
			if(devfd < 0) 
				continue; 

			if(ioctl(devfd, B_GET_GEOMETRY, &g, sizeof(g)) >= 0) {
				if(g.device_type == B_CD)
				{ 
					return devfd;
				}
			}
			close(devfd);
		} 
	}
	return B_ERROR;
}

int
CDEngine::FindCDPlayerDevice()
{
	return try_dir("/dev/disk");
}



